import logging
import os
import re
from pathlib import Path
from typing import Union
from urllib.error import HTTPError
from urllib.parse import urlparse
from typing import Callable

import pandas as pd
import requests
from O365.drive import File
from bs4 import BeautifulSoup as bs

from rypython.ry365 import WorkBook

logging.basicConfig(level=logging.INFO)

HEADER = {
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.75 Safari/537.36",
    "X-Requested-With": "XMLHttpRequest"
}

DEFAULT_DOWNLOAD_DIR = os.environ.get(
    'RYPYTHON_DEFAULT_DOWNLOAD_DIR',
    Path.home() / 'Downloads'
)


def read_excel365(
        xl_file: Union[File, WorkBook],
        sheet_name: str = False,
        skip_rows: int = 0
):
    wb = WorkBook(xl_file) if isinstance(xl_file, File) else xl_file
    if sheet_name is None:
        wss = wb.get_worksheets()
        dfs = {}
        for ws in wss:
            dfs[ws.name] = read_excel365(wb, sheet_name=ws.name, skip_rows=skip_rows)
        return dfs
    if sheet_name is False:
        wss = wb.get_worksheets()
        ws = wss[0]
    elif sheet_name:
        ws = wb.get_worksheet(sheet_name)
    _range = ws.get_used_range()
    cols, *values = _range.values[skip_rows:]
    df = pd.DataFrame(values, columns=cols)
    df.range = _range
    return df


class HTMLDataFrame:
    def __init__(self, url: str):
        self.url = urlparse(url)
        self.tables = self._parse_url() or []

    def _parse_url(self, text: str = None):
        try:
            return pd.read_html(text or self.url.geturl())
        except AttributeError as e:
            logging.error(f"Error parsing {self.url.geturl()}: {e}!")
            r = requests.get(self.url.geturl(), headers=HEADER)
            text = r.text
            rowspan_pattern = r'(?P<rowspan>rowspan="\d+)(&([a-z])+;)+"'
            text = re.sub(rowspan_pattern, r'\g<rowspan>"', text)
            colspan_pattern = r'(?P<colspan>colspan="\d+)(&([a-z])+;)+"'
            text = re.sub(colspan_pattern, r'\g<colspan>"', text)
            return pd.read_html(text)
        except HTTPError as e:
            logging.error(f"Error parsing {self.url.geturl()}: {e}!")
            r = requests.get(self.url.geturl(), headers=HEADER)
            self._parse_url(text=r.text)
        except ValueError as e:
            logging.error(f"Error parsing tables for {self.url.geturl()}: {e}!")
            return []

    def get_list_items(self):
        items = {}
        r = requests.get(self.url.geturl(), headers=HEADER)
        soup = bs(r.text, 'lxml')
        uls = soup.find_all('ul')
        for ul in uls:
            for li in ul.find_all('li'):
                if item := li.text:
                    items.setdefault('Bulleted Lists', []).append(item)
        ols = soup.find_all('ol')
        for ol in ols:
            for li in ol.find_all('li'):
                if item := li.text:
                    items.setdefault('Ordered Lists', []).append(item)
        if items:
            return {
                list_type: pd.DataFrame(list_items, columns=['List Items'])
                for list_type, list_items in items.items()
            }
        return items

    def export(
            self,
            output_dir: Path = None,
            filename: str = None,
            transform_df: Callable[[pd.DataFrame], pd.DataFrame] = None
    ):
        output_dir = output_dir or DEFAULT_DOWNLOAD_DIR
        filename = filename or self.url.path.replace('/', '_')
        if not filename.endswith('.xlsx'):
            filename = f"{filename}.xlsx"
        outfile = output_dir / filename
        table_count = 1
        with pd.ExcelWriter(outfile, engine='xlsxwriter') as writer:
            source_df = pd.DataFrame(
                [[self.url.geturl()]],
                columns=['Source']
            )
            source_df.to_excel(
                writer,
                sheet_name='Source',
                index=False
            )
            for table in self.tables:
                first_column = table.columns[0]
                if (not isinstance(first_column, str)) or all(
                        l.isdigit()
                        for l in first_column
                ):
                    table.columns = table.iloc[0]
                    table = table.drop(table.index[0])
                if table.empty:
                    continue
                if transform_df is not None:
                    table = transform_df(table)
                table.to_excel(
                    writer,
                    sheet_name=f"Table {table_count}",
                    index=False
                )
                table_count += 1
            for list_type, list_df in self.get_list_items().items():
                list_df.to_excel(
                    writer,
                    sheet_name=list_type,
                    index=False
                )


class DataFrame:
    """
    The ``DataFrame`` class provides a quick process for
    downloading and extracting information from Excel files on SharePoint.
    Pulling the data directly from the files is possible, but generally very slow.
    By downloading the file, it can be loaded directly with ``pandas``.

    Parameters
    ----------
    source_file: File
    download_dir: Path
    sheet_name: str
    **options: dict

    Attributes
    ----------
    source_file: File
    download_dir: Path
    sheet_name: str
    **options: dict

    Examples
    --------
    ``DataFrame`` is designed for use as a context manager.
    >>> with DataFrame(
    >>>        source_file,
    >>>        download_dir,
    >>>        sheet_name,
    >>>        **options
    >>>        ) as wb:
    >>>            wb.do_stuff_with_workbook()
    """
    def __init__(
            self,
            source_file: File,
            download_dir: Path = DEFAULT_DOWNLOAD_DIR,
            sheet_name: str = None,
            **options
    ):
        self.source_file = source_file
        self.file_type = source_file.name.rsplit('.', 1)[1].lower()
        self.download_dir = download_dir
        self.sheet_name = sheet_name
        self.options = options

    def __enter__(self):
        self.source_file.download(self.download_dir)
        self.local_file = self.download_dir / self.source_file.name
        self.data = self._parse_source_file()
        return self

    def _parse_source_file(self):
        if self.file_type == 'csv':
            return pd.read_csv(
                self.local_file,
                **self.options
            )
        return pd.read_excel(
            self.local_file,
            sheet_name=self.sheet_name,
            **self.options
        )

    def __exit__(self, type, value, traceback):
        os.remove(self.local_file)



import json
import logging
from dataclasses import dataclass
from typing import List, Union, Dict, Any

import numpy as np
import pandas as pd
from pydantic import BaseModel
from rich.table import Table

from rypython.rex import capture_all
from rypython.ryagram import erDiagram

logging.basicConfig(level=logging.DEBUG)


@dataclass
class DType:
    pandas: str
    sql: str = None


class DataType:
    MAPPING = {
        "string": {
            "tiny": "TINYTEXT",
            "text": "TEXT",
            "med": "MEDIUMTEXT",
            "long": "LONGTEXT"
        },
        "int": {
            "int8": "TINYINT",
            "int16": "SMALLINT",
            "int32": "INT",
            "int64": "BIGINT"
        },
        "float": "FLOAT",
        "date": "DATE",
        "time": "TIME",
        "datetime": "DATETIME"
    }

    def __init__(self, column: pd.Series):
        self.pandas = column.dtype
        self.sql = self.to_sql_dtype(column)

    @staticmethod
    def category_to_sql_dtype(column: pd.Series):
        categories = column.astype("category").dtype.categories.to_list()
        return f"ENUM({', '.join(category for category in categories)})"

    @staticmethod
    def str_to_sql_char(char_lens: np.ndarray):
        if char_lens.shape[0] == 1:
            return f"CHAR({char_lens[0]})"

    @staticmethod
    def str_to_sql_dtype(column: pd.Series):
        string_mapping = DataType.MAPPING.get("string")
        max_len = column.map(len).max()

        if (
                (char_lens := column.map(len).unique())
        ):
            if sql_dtype := DataType.str_to_sql_char(char_lens):
                return sql_dtype

        if max_len < pow(2, 8):
            return string_mapping.get("tiny")
        if max_len < pow(2, 16):
            return string_mapping.get("text")
        if max_len < pow(2, 24):
            return string_mapping.get("med")
        return string_mapping.get("long")

    @staticmethod
    def int_to_sql_dtype(column: pd.Series):
        int_mapping = DataType.MAPPING.get("int")
        dtype = column.dtype.lower()

        if sql_dtype := int_mapping.get(dtype):
            return sql_dtype

        max_value = column.max()
        if max_value < pow(2, 16):
            return int_mapping.get("int16")
        if max_value < pow(2, 32):
            return int_mapping.get("int32")
        return int_mapping.get("int64")

    @staticmethod
    def to_sql_dtype(
            column: pd.Series = Any
    ):
        dtype = column.dtype.lower()
        sql_mapping = DataType.MAPPING.get(dtype)
        if isinstance(sql_mapping, str):
            return sql_mapping
        if dtype == "category":
            return DataType.category_to_sql_dtype(column)
        if "int" in dtype:
            if (
                    (len_counts := column.astype("string").map(len).unique().shape[0]) == 1
            ):
                return DataType.to_sql_type(column.astype("string"))
            return DataType.int_to_sql_dtype(column)
        if dtype == "string":
            return DataType.int_to_sql_dtype(column)


class RyDBColumn:
    def __init__(self, column_name: str, column: pd.Series):
        self.name = column_name
        self.column = column
        self.dtype = DataType(column)


class CalculatedColumn(RyDBColumn):
    def __init__(
            self,
            column_name: str,
            source: Union[pd.Series, pd.DataFrame],
            source_label: str = None,
            **config
    ):
        column = self._transform(source, config)
        super().__init__(column_name, column)
        self.source_label = source_label

    def _transform(self, source: Union[pd.Series, pd.DataFrame], config: Dict[str, Any]):
        pass


class RyDBTable:
    def __init__(self, table_name: str, table: pd.DataFrame):
        self.name = table_name
        self.df = table
        self._refresh()

    def to_mermaid(self):
        diagram = erDiagram()
        diagram.add_entity_by_dtypes(
            self.name,
            self.df.dtypes
        )

    @staticmethod
    def _extract_attribute_from_base_model(base_model: BaseModel, attr_name: str):
        if not attr_name.startswith("$"):
            return getattr(base_model, attr_name)
        attr_path = attr_name.replace("$", "").split(".")
        attr_value = base_model
        for attr_level in attr_path:
            attr_value = getattr(attr_value, attr_level)
        return attr_value

    @staticmethod
    def _transform_value(value: Any, transform_type: str):
        pass

    @classmethod
    def from_pydantic(
            cls,
            table_name: str,
            table_def: Dict[str, str],
            column_mappings: List[Dict[str, str]],
            *base_models: BaseModel,
            ignore_by_any: Any = None,
            ignore_by_path: Any = None
    ):
        table_values = []
        for base_model in base_models:
            row_values = []
            for column_name, dtype in table_def.items():
                logging.debug(f"Calculating {column_name}:{dtype}")
                column_mapping = column_mappings.get(column_name)

                attr_name = column_mapping.get("attr_name")
                attr_value = cls._extract_attribute_from_base_model(
                    base_model,
                    attr_name
                )
                if transform_config := column_mapping.get("transform", {}):
                    attr_value = cls._transform_value(
                        attr_value,
                        **transform_config
                    )
                row_values.append(attr_value)
        table = pd.DataFrame(
            table_values,
            columns=table_def.keys().tolist(),
            dtype=table_def
        )
        return cls(table_name, table)

    def __repr__(self):
        return repr(self.df)

    @property
    def shape(self):
        return self.df.shape

    @property
    def columns(self):
        return self.df.columns.tolist()

    def loc(
            self,
            query: Union[str, pd.Series],
            column_names: List[str] = None,
            squeeze: bool = False
    ):
        mask = self.df.eval(query) if isinstance(query, str) else query
        target = (mask,) if column_names is None else (mask, column_names)
        result = self.df.loc[target]
        return result.squeeze() if squeeze else result

    @staticmethod
    def format_where_query(where_dict: dict):
        where_conds = set()
        for where_column, where_value in where_dict.items():
            where_cond = f"{where_column} == {where_value}"
            where_conds.add(where_cond)
        if len(where_conds) == 1:
            return list(where_conds)[0]
        where_conds = [
            f"({where_cond})"
            for where_cond in where_conds
        ]
        return " & ".join(where_conds)

    def stringify(self, *args):
        return [
            str(arg)
            for arg in args
        ]

    def update_row(
            self,
            set_dict: dict,
            where_dict: dict
    ):
        set_columns = list(set_dict.keys())
        set_values = list(set_dict.values())
        where_query = self.format_where_query(where_dict)
        self.df.loc[self.df.eval(where_query), set_columns] = set_values
        logging.info(f"({','.join(self.stringify(set_columns))}) updated to ({','.join(self.stringify(set_values))}) for {where_query}")

    def lookup(self, *args):
        return self.loc(*args)

    def _refresh(self):
        self.__dict__.update(
            {
                column_name: self.df[column_name]
                for column_name in self.columns
            }
        )

    @staticmethod
    def regextract(pattern: str, value: str):
        return json.dumps(
            capture_all(
                pattern,
                value,
                flatten=True
            )
        )

    def add_columns_by_regex(
            self,
            pattern: str,
            source_column_name: str,
            column_order: List[str] = None,
            column_mappings: dict = None,
            boolean_columns: List[str] = None
    ) -> None:
        source_column = getattr(self, source_column_name)
        json_column_values = source_column.apply(
            lambda x: self.regextract(pattern, x)
        ).apply(json.loads).tolist()
        json_columns = pd.DataFrame(json_column_values)
        for json_column in json_columns:
            if column_mapping := column_mappings.get(json_column):
                json_columns[json_column] = json_columns[json_column].apply(
                    lambda x: column_mapping.get(x)
                )
            if json_column in boolean_columns:
                json_columns[json_column] = np.where(
                    json_columns[json_column].notna(),
                    True,
                    False
                )
        if column_order is not None:
            json_columns = json_columns[column_order]
        for json_column in json_columns:
            self.df[json_column] = json_columns[json_column]
        self._refresh()

    def append_right(self, right_df: pd.DataFrame):
        self.df = self.df.join(right_df)
        self._refresh()

    def to_rich_console_table(
            self,
            query: str,
            name: str = None,
            column_formats: dict = None,
            count: int = 10,
            fillna: str = ""
    ):
        table = Table(title=name or self.name)
        results = self.df.query(query).head(count)

        columns = self.columns if column_formats is None else list(column_formats.keys())

        for column in columns:
            column_format = column_formats.get(column, {})
            table.add_column(column, **column_format)

        for row in results.itertuples():
            row_values = [
                getattr(row, column, fillna)
                for column in columns
            ]
            table.add_row(*row_values)

        return table



from typing import List, Callable

import pendulum as pm
import pymsteams as teams


class TeamsAlert:
    def __init__(self, alert_title: str, bot_link: str):
        self.title = alert_title
        self.bot_link = bot_link

    def __call__(self, func):

        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                message = teams.connectorcard(self.bot_link)
                message.title(self.title)
                message.text(f"{e} ({pm.now():%Y-%m-%d %H:%M:%S})")
                message.send()

        return wrapper


class AdaptiveTextBlock:
    def __init__(
            self,
            text: str,
            wrap: bool = True,
            weight: str = 'Bolder'
    ):
        self.text = text
        self.wrap = wrap
        self.weight = weight

    @property
    def json(self):
        return {
            "type": "TextBlock",
            "text": self.text,
            "wrap": self.wrap,
            "weight": self.weight
        }


class AdaptiveContainer:
    def __init__(self, items: List[dict]):
        self.items = items

    @property
    def json(self):
        return {
            "type": "Container",
            "items": self.items,
            "separator": True
        }


class AdaptiveBulletedList:
    def __init__(
            self,
            items: List[str],
            header: AdaptiveTextBlock = None,
            tfunc: Callable[[str], str] = None,
            wrap: bool = True,
            spacing: str = 'Small',
            weight: str = 'Default'
    ):
        self.header = header
        self.items = [
            tfunc(item)
            for item in items
        ] if tfunc is not None else items
        self.wrap = wrap
        self.spacing = spacing
        self.weight = weight

    @property
    def item_json(self):
        return {
            "type": "TextBlock",
            "text": ' \r'.join(
                f"- {item}"
                for item in self.items
            ),
            "wrap": self.wrap,
            "spacing": self.spacing,
            "weight": self.weight
        }

    @property
    def json(self):
        items = [
            self.header.json,
            self.item_json
        ] if self.header is not None else [
            self.item_json
        ]
        container = AdaptiveContainer(items)
        return container.json


class AdaptiveCard:
    def __init__(
            self,
            bot_link: str,
            title: str,
            size: str = 'ExtraLarge',
            weight: str = 'Bolder'
    ):
        self.message = teams.connectorcard(bot_link)
        self.title = title
        self.size = size
        self.weight = weight
        self.body = []

    @property
    def json(self):
        return {
            "type": "AdaptiveCard",
            "body": [
                {
                    "type": "TextBlock",
                    "size": self.size,
                    "weight": self.weight,
                    "text": self.title
                },
                *[
                    element.json
                    for element in self.body
                ]
            ],
            "$schema": "http://adaptivecards.io/schemas/adaptive-card.json",
            "version": "1.0"
        }

    def add_text_block(
            self,
            text: str,
            wrap: bool = True,
            weight: str = 'Bolder'
    ):
        self.body.append(
            AdaptiveTextBlock(text, wrap, weight)
        )

    def add_bulleted_list(
            self,
            items: List[str],
            header: AdaptiveTextBlock = None,
            tfunc: Callable[[str], str] = None,
            wrap: bool = True,
            spacing: str = 'Small',
            weight: str = 'Default'
    ):
        self.body.append(
            AdaptiveBulletedList(
                items,
                header,
                tfunc,
                wrap,
                spacing,
                weight
            )
        )

    @property
    def payload(self):
        return {
            'type': 'message',
            'attachments': [
                {
                    'contentType': 'application/vnd.microsoft.card.adaptive',
                    'content': self.json
                }
            ]
        }

    def send(self):
        self.message.payload = self.payload
        self.message.send()


import re

def stringify(q: str):
    return q if isinstance(q, str) else str(q)

def capture_all(
        pattern: str,
        query: str,
        allow_empty: bool = True,
        flatten: bool = False
):
    query = stringify(query)
    r = re.compile(pattern)
    s = [
        m.groupdict()
        for m in r.finditer(query)
    ]
    d = {}
    for match in s:
        for key, value in match.items():
            if not value and not allow_empty:
                continue
            d.setdefault(key, []).append(value)
    if flatten:
        for key, value in d.items():
            if isinstance(value, list):
                if len(value) == 1:
                    d[key] = value[0]
                if len(value) == 0:
                    d[key] = None
    return d

def is_format(pattern: str, query: str):
    query = stringify(query)
    r = re.compile(pattern)
    return bool(re.match(r, query))


import logging
import os
from pathlib import Path
from typing import List, Union, Dict, Any

import pandas as pd
import pyodbc

from rypython.randas import DataFrame
from rypython.ry365 import O365Account
from rypython.rydb.tables import RyDBTable

logging.basicConfig(level=logging.DEBUG)

DEFAULT_DOWNLOAD_DIR = os.environ.get(
    'RYPYTHON_DEFAULT_DOWNLOAD_DIR',
    Path.home() / 'Downloads'
)


class RyDBSource:
    def __init__(self, type: str, **kwargs):
        self.type = type
        self.updated = set()

    @property
    def has_changed(self):
        return bool(self.updated)

    def connect(self):
        ...

    def commit(self):
        ...

    def collect_table(self, table_name: str):
        return self.conn.get(table_name)

    def collect_all(self):
        ...

    def replace(self, tables: Dict[str, pd.DataFrame]) -> None:
        ...

    def update_row(
            self,
            table_name: RyDBTable,
            set_dict: dict,
            where_dict: dict
    ):
        # table = self.collect_table(table_name)
        self.updated.add(table_name)
        self.conn[table_name].update_row(set_dict=set_dict, where_dict=where_dict)


class O365DB(RyDBSource):
    def __init__(
            self,
            site: str,
            filepath: str,
            db_filename: str
    ):
        super().__init__(type="o365")
        self.site = site
        self.filepath = "/".split(filepath)
        self.db_filename = db_filename
        self.db_file = self.connect()
        self.conn = self.collect_all()

    def connect(self):
        account = O365Account(site=self.site)
        folder = account.get_folder(*self.filepath)
        return folder.get_item(self.db_filename)

    def collect_table(self, table_name):
        with DataFrame(self.db_file, sheet_name=table_name) as db:
            return RyDBTable(table_name, db.data)


    def collect_all(self):
        with DataFrame(self.db_file) as db:
            return {
                table_name: RyDBTable(table_name, table)
                for table_name, table in db.data.items()
            }

    def commit(self):
        if self.has_changed:
            local_db_file = DEFAULT_DOWNLOAD_DIR / self.db_file.name
            with pd.ExcelWriter(local_db_file, engine="xlsxwriter") as writer:
                for table_name, table in self.conn.items():
                    table.to_excel(writer, sheet_name=table_name, index=False)
                if not local_db_file.exists():
                    logging.error(f"Could not find {local_db_file}!")
                    return
                folder = self.db_file.get_parent()
                self.db_file.delete()
                folder.upload_file(local_db_file)
                if new_db_file := folder.get_item(local_db_file.name):
                    logging.info(f"Database file updated at {new_db_file}")

    def list_tables(self):
        return list(self.data.keys())
    def update_row(
            self,
            table_name: str,
            set_dict: dict,
            where_dict: dict
    ):
        super().update_row(table_name, set_dict=set_dict, where_dict=where_dict)

    def replace(self, tables: Dict[str, pd.DataFrame]) -> None:
        folder = self.folder
        existing_db_file = folder.get_item(self.db_filename)
        new_local_db_file = DEFAULT_DOWNLOAD_DIR / self.db_filename
        with pd.ExcelWriter(
                new_local_db_file,
                engine="xlsxwriter"
        ) as writer:
            for table_name, table in tables.items():
                if table_name == "source":
                    continue
                table.to_excel(writer, sheet_name=table_name, index=False)
        if new_local_db_file.exists() and existing_db_file is not None:
            existing_db_file.delete()
            self.folder.upload_file(new_local_db_file)


class SQLDB(RyDBSource):
    def __init__(
            self,
            username: str,
            password: str,
            server: str,
            database: str,
            port: int = 1433,
            driver: str = "{ODBC Driver 17 for SQL Server}",
            engine: str = "pyodbc"
    ):
        super().__init__(type="sql")
        self.server = server
        self.host = server.split(".", 1)[0]
        self.database = database
        self.port = port
        self.driver = driver
        self.engine = engine
        self._conn = self.connect(username, password)
        self.conn = self.collect_all()
        self.curr = self._conn.cursor()

    def connect(self, username: str, password: str):
        if self.engine != "pyodbc":
            return
        conn_string = ";".join(
            [
                f"DRIVER={self.driver}",
                f"PORT={self.port}",
                f"SERVER={self.server}",
                f"DATABASE={self.database}",
                f"UID={username}",
                f"PWD={password}"
            ]
        )
        return pyodbc.connect(conn_string)

    def collect_table(self, table_name: str = None):
        query = f"""
        SELECT * FROM {table_name} 
        """
        return RyDBTable(table_name, pd.read_sql(query, self._conn))

    def collect_all(self):
        table_names = self.list_tables()
        return {
            table_name: self.collect_table(table_name)
            for table_name in table_names
        }

    def list_tables(self):
        query = """
        SELECT schema_name(t.schema_id) as schema_name,
            t.name as table_name,
            t.create_date,
            t.modify_date
        FROM sys.tables t
        ORDER BY schema_name, table_name;
        """
        df = pd.read_sql(query, self._conn)
        return df.table_name.tolist()

    def commit(self):
        if self.has_changed:
            self._conn.commit()

    def execute(
            self,
            query: str,
            *values: Any
    ):
        logging.info(f"Sending {query} with {values}")
        self.curr.execute(
            query,
            *values
        )

    def update_row(
            self,
            table_name: str,
            set_dict: dict,
            where_dict: dict,
            _execute: bool = True,
            _replace: bool = False
    ):
        super().update_row(table_name, set_dict=set_dict, where_dict=where_dict)
        if _execute:
            logging.info("Updating server table")
            set_clause = ", ".join(f"{set_key} = ?" for set_key in set_dict)
            where_clause = ", ".join(f"{where_key} = ?" for where_key in where_dict)
            values = [
                *set_dict.values(),
                *where_dict.values()
            ]
            query = f"""
                    UPDATE
                        {table_name}
                        SET {set_clause}
                        WHERE {where_clause}
                    """
            self.execute(query, *values)
        if _replace:
            logging.info("Print replacing server table")
            table = self.conn[table_name]
            table.df.to_sql(table_name, con=self._conn, if_exists="replace")







class RyDB:
    def __init__(
            self,
            source: RyDBSource,
            read_only: bool = False,
            collect_all: bool = True
    ):
        self.source = source
        self.read_only = read_only
        if collect_all:
            self.source.collect_all()
        self.conn = self.source.conn

    def __setattr__(self, table_name: str, table: pd.DataFrame):
        super().__setattr__(table_name, table)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if not self.read_only:
            self.commit()

    def commit(self):
        self.source.commit()

    def collect_table(self, table_name: str):
        return self.source.collect_table(table_name)

    def collect_all(self):
        return self.source.collect_all()

    @classmethod
    def read_o365(
            cls,
            site: str,
            filepath: Union[str, List[str]],
            db_filename: str
    ):
        SOURCE = O365DB(
            site=site,
            filepath=filepath,
            db_filename=db_filename
        )
        return cls(source=SOURCE)

    @classmethod
    def read_sql(
            cls,
            server: str,
            username: str,
            password: str,
            database: str,
            port: str = 1433,
            driver: str = "{ODBC Driver 17 for SQL Server}",
            engine: str = "pyodbc"
    ):
        SOURCE = SQLDB(
            username,
            password,
            server=server,
            database=database,
            port=port,
            driver=driver,
            engine=engine
        )
        return cls(source=SOURCE)

    def replace(self):
        self.source.replace(
            {
                table_name: table.df
                for table_name, table in self.__dict__.items()
                if isinstance(table, RyDBTable)
            }
        )

    def count(self, table_name: str, column_name: str = None):
        return

    def update_row(self, table_name: str, set_dict: dict, where_dict: dict):
        self.source.update_row(table_name, set_dict, where_dict)



import PySimpleGUI as sg
from typing import List, Any


class NewTab:
    def __init__(self, tab_name: str, tab_layout: List[List[Any]]):
        self.tab = sg.Tab(
            tab_name,
            tab_layout
        )

    @property
    def config(self):
        return [
            self.tab
        ]


class NewTabGroup:
    def __init__(self, tabs: List[NewTab]):
        self.tab_group = sg.TabGroup(
            [
                tab.config
                for tab in tabs
            ]
        )

    @property
    def config(self):
        return [
            [self.tab_group]
        ]



class NewWindow:
    def __init__(self, name, layout):
        self.event = None
        self.values = None
        self.name = name
        self.layout = layout

    def __enter__(self):
        self.window = sg.Window(self.name, self.layout)
        return self

    def __exit__(self, type, value, traceback):
        self.window.close()

    def read(self):
        self.event, self.values = self.window.read()






from __future__ import annotations

import pandas as pd
from typing import Dict, List, Any, Union
from dataclasses import dataclass

class Mermaid:
    OPEN: str = "```mermaid"
    CLOSE: str = "```"
    INDENT: str = "    "

    def __init__(self, declaration: str):
        self.declaration = declaration
        self.lines = []

    def add_line(self, new_line: str):
        self.lines.append(new_line)

    def print(self):
        lines = [
            self.OPEN,
            self.declaration,
            *self.lines,
            self.CLOSE
        ]
        print("\n".join(lines))


@dataclass
class Section:
    name: str

    @property
    def declaration(self):
        return f"    section {self.name}"


@dataclass
class Task:
    name: str
    description: str
    turnaround: str
    is_critical: bool | str = False
    after: List[str] = None
    start_date: str = ""
    end_date: str = ""

    @property
    def is_active(self):
        return (bool(self.start_date) and not bool(self.end_date))

    @property
    def is_milestone(self):
        if self.name.lower() == "milestone":
            return True
        return False

    @property
    def is_done(self):
        if self.is_milestone:
            return False
        return bool(self.end_date)

    @property
    def declaration(self):
        if self.is_milestone:
            return f"    {self.description}\t:milestone, {self.turnaround}, 0d"
        crit = "crit, " if self.is_critical else ""
        status = ""
        if self.is_active:
            status = "active, "
        if self.is_done:
            status = "done, "
        after = f" after {' '.join(self.after)}," if self.after is not None else ""
        end_date = self.end_date or self.turnaround
        return f"    {self.description}\t:{crit}{status}{self.name},{after} {self.start_date}, {end_date}"




class ganttChart(Mermaid):
    def __init__(
            self,
            title: str,
            end_date: str,
            date_format: str = "YYYY-MM-DD",
            excludes: List[str] = None,
            note: str = None
    ):
        super().__init__(declaration="gantt")
        self.title = title
        self.end_date = end_date
        self.date_format = date_format
        self.excludes = ", ".join(excludes) if excludes else ""
        self.note = note
        self._add_description_lines()

    def _add_description_lines(self):
        date_format = f"{self.INDENT}dateFormat\t{self.date_format}"
        self.lines.append(date_format)
        title = f"{self.INDENT}title\t{self.title}"
        self.lines.append(title)
        excludes = f"{self.INDENT}excludes\t{self.excludes}"
        self.lines.append(excludes)
        if self.note is not None:
            note = f"{self.INDENT}%% {self.note}"
            self.lines.append(note)

    def add_section(
            self,
            name: str,
            tasks: List[Task]
    ):
        self.lines.append(self.INDENT)
        section = Section(name=name)
        self.lines.append(section.declaration)
        for task in tasks:
            self.lines.append(task.declaration)






class classDiagram(Mermaid):
    def __init__(self):
        super().__init__(declaration="classDiagram")

    def add_class(
            self,
            class_instance: Any,
            ignore_private: bool = True
    ):
        class_name = class_instance.__class__.__name__
        self.lines.append(f"{self.INDENT} class {class_name} {{")
        attr_dict = {}
        sorted_attr = {}
        all_attrs = dir(class_instance)
        for attr_name in all_attrs:
            try:
                attr_dict[attr_name] = getattr(class_instance, attr_name)
            except AttributeError:
                continue

        for attr_name, attr in attr_dict.items():
            if attr_name.startswith("_") and ignore_private:
                continue
            if attr_name in class_instance.__dict__:
                sorted_attr.setdefault(
                    "attributes", []
                ).append((attr_name, type(attr).__name__))
                continue
            if f"bound method {class_name}" in repr(attr):
                sorted_attr.setdefault(
                    "methods", []
                ).append(attr_name)

        attributes = sorted(sorted_attr.get("attributes"))
        for attr_name, attr_type in attributes:
            self.lines.append(f"{self.INDENT*2} +{attr_type} {attr_name}")

        methods = sorted(sorted_attr.get("methods"))
        for method_name in methods:
            self.lines.append(f"{self.INDENT * 2} +{method_name}()")
        self.add_line(f"{self.INDENT}}}")







class erDiagram(Mermaid):
    def __init__(self):
        super().__init__(declaration="erDiagram")

    def add_entity(
            self,
            entity_name: str,
            dtypes: Dict[str, str],
            primary_key: str = None,
            foreign_keys: List[str] = None,
            descriptions: Dict[str, str] = None
    ):
        foreign_keys = foreign_keys or []
        descriptions = descriptions or {}
        self.add_line(f"{self.INDENT}{entity_name} {{")
        for column_name, dtype in dtypes.items():
            base_line = f"{self.INDENT*2}{dtype} {column_name}"
            key_status = ""
            if column_name == primary_key:
                key_status = "PK"
            if column_name in foreign_keys:
                key_status = "FK"
            description = descriptions.get(column_name, "")
            full_line = f"{base_line} {key_status} {description}"
            self.add_line(full_line)
        self.add_line(f"{self.INDENT}}}")



    def add_entity_by_dtypes(
            self,
            entity_name: str,
            dtypes: pd.Series,
            primary_key: str = None,
            foreign_keys: List[str] = None,
            descriptions: Dict[str, str] = None
    ):
        dtypes = {
            column_name: dtype.name
            for column_name, dtype in dtypes.copy().reset_index().values.tolist()
        }
        self.add_entity(
            entity_name,
            dtypes,
            primary_key=primary_key,
            foreign_keys=foreign_keys,
            descriptions=descriptions
        )




import string
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, Callable, List, Any, Union, Generator, Dict

import numpy as np
import pandas as pd
from xlsxwriter import Workbook
from xlsxwriter.format import Format
from xlsxwriter.worksheet import Worksheet

DataFrameRow = Any  # TODO: Figure out how to type hint Pandas rows
DataFrameIndex = Any


@dataclass
class SheetPosition:
    start: str

    def __post_init__(self):
        self.row = int(''.join(l for l in self.start if l.isdigit()))
        self.columns = string.ascii_uppercase
        column = self.start.replace(str(self.row), "")
        self.column = self.columns.index(column)
        self.left_column = self.column
        self.right_column = self.left_column - 1

    def up(self):
        self.row -= 1

    def down(
            self,
            return_left: bool = True,
            return_right: bool = False
    ):
        self.row += 1
        if return_left:
            self.return_left()
        if return_right:
            self.return_right()

    def right(self):
        self.column += 1
        self.right_column = max(self.column, self.right_column) - 1

    def left(self):
        self.column -= 1

    def return_left(self):
        self.column = self.left_column

    def return_right(self):
        self.column = self.right_column

    def current_merge_range(self, width: int = None):
        left_column = self.columns[self.left_column]
        right_index = (self.left_column + width - 1) if width is not None else self.right_column
        right_column = self.columns[right_index]
        return f"{left_column}{self.row}:{right_column}{self.row}"


    @property
    def pos(self):
        column = self.columns[self.column]
        return f"{column}{self.row}"


@dataclass
class RexcelFormat:
    test_func: Callable
    config: dict
    format_type: str = "row"  # TODO: Set up as ``Enum``

    def get_format(
            self,
            check_value: Any
    ):
        ...


class RexcelWorksheet:
    """
    Worksheet object for adding formatted DataFrames to Excel

    Parameters
    ----------
    workbook: Workbook
        ``Workbook`` object used to add worksheets
    worksheet_name: str
        Name of new Excel worksheet
    column_widths: List[Tuple[int, int, int]]
        List of (start, stop, width) tuples for assigning column widths in the sheet

    Attributes
    ----------
    workbook: Workbook
        ``Workbook`` object used to add worksheets
    wks: Worksheet
        New worksheet as ``Worksheet`` object
    column_widths: List[Tuple[int, int, int]]
        List of (start, stop, width) tuples for assigning column widths in the sheet
    """

    def __init__(
            self,
            workbook: Workbook,
            worksheet_name: str,
            column_widths: List[Tuple[int, int, int]] = None,
            image_config: dict = None
    ) -> None:
        self.workbook = workbook
        self.wks = self.workbook.add_worksheet(
            name=worksheet_name
        )

        if column_widths:
            for start, stop, width in column_widths:
                self.wks.set_column(
                    start,
                    stop,
                    width
                )

        if image_config is not None:
            self.insert_image(**image_config)

        self.current_row = 1
        self.current_column = 0

        self.data_validation_columns = {}

        # TODO: Improve this implementation of Formats
        self.FORMATS = {
            'bold': self.workbook.add_format({'bold': 1}),
            'text': self.workbook.add_format({'num_format': '@'}),
            'percent': self.workbook.add_format({'num_format': 9}),
            'integer': self.workbook.add_format({'num_format': 1}),
            'decimal': self.workbook.add_format({'num_format': 2}),
            'locked': self.workbook.add_format({'locked': True})
        }

    @staticmethod
    def get_column_generator(
            limit: int = 26,
            row_number: int = None
    ) -> Generator[Tuple[str, int], None, None]:
        """
        Creates generator object for identifying columns

        Parameters
        ----------
        limit: int
        row_number: int

        Returns
        -------
        Generator[Tuple[str, int], None, None]
            Returns generator that yields (column_letter, column_number, cell_address) tuple
        """
        BASE_COLUMNS = list(string.ascii_uppercase)
        row_number = str(row_number) if row_number else ""
        prefix_offset = -1
        column_offset = 0
        column_count = 0
        while column_count < limit:
            if column_offset == 26:
                column_offset = 0
                prefix_offset += 1
            letter_prefix = BASE_COLUMNS[prefix_offset] if prefix_offset >= 0 else ""
            base_letter = BASE_COLUMNS[column_offset]
            column_letter = f"{letter_prefix}{base_letter}"
            cell_address = f"{column_letter}{row_number}"
            yield column_letter, column_count, cell_address
            column_offset += 1
            column_count += 1

    def get_write_func(self, cell_value: Any, dtype: str = None):
        if dtype is None:
            if isinstance(cell_value, (int, float)):
                return self.wks.write_number
            if isinstance(cell_value, bool):
                return self.wks.write_boolean
            if cell_value == "":
                return self.wks.write_blank
            if cell_value.startswith("="):
                return self.wks.write_formula
            return self.wks.write_string

    def simple_write_df(
            self,
            df: pd.DataFrame,
            starting_cell: str,
            header_format: str = "bold",
            subtitle: str = None,
            subtotal: str = None,
            apply_columns: bool = True
    ):
        current = SheetPosition(starting_cell) if isinstance(starting_cell, str) else starting_cell
        columns = df.columns.tolist()
        if apply_columns:
            for column_header in columns:
                self.write_cell(
                    current.pos,
                    'write_string',
                    column_header,
                    header_format
                )
                current.right()
            current.down()
        if subtitle is not None:
            merge_range = current.current_merge_range(width=df.shape[1])
            merge_format = self.workbook.add_format(
                {
                    "bold": 1,
                    "border": 1,
                    "align": "center",
                    "valign": "vcenter",
                    "bg_color": "#FFDAC4",
                    "font_color": "#63666A",
                    "border_color": "#FF6A13"
                }
            )
            self.wks.merge_range(
                merge_range,
                subtitle,
                merge_format
            )
            current.down()
        subtotal_cells = []
        for row in df.fillna("").itertuples():
            for i, cell_value in enumerate(row):
                if i == 0:
                    continue
                if columns[i-1] == subtotal:
                    subtotal_cells.append(current.pos)
                write_func = self.get_write_func(cell_value)
                self.write_cell(
                    current.pos,
                    write_func,
                    cell_value
                )
                current.right()
            current.down()
        if subtotal is not None:
            current.return_right()
            current.left()
            self.write_cell(
                current.pos,
                self.wks.write_string,
                "subtotal",
                header_format
            )
            current.right()
            subtotal_range = f"{subtotal_cells[0]}:{subtotal_cells[-1]}"
            self.write_cell(
                current.pos,
                self.wks.write_formula,
                f"=SUM({subtotal_range})"
            )
        return current


    def write_cell(
            self,
            cell: str,
            write_func: Union[str, Callable],
            cell_text: str,
            cell_format: str = None
    ) -> None:
        """
        Writes given text to specified cell according to write function and cell format

        Parameters
        ----------
        cell: str
            Cell address in "A1" format
        write_func: Union[str, Callable]
            Write function to call ("write_string", "write_formula", "write_boolean", "write_number")
        cell_text: str
            Text to write
        cell_format: str
            Format for cell ("bold", "text", etc.)
        """
        cell_format = self.FORMATS.get(
            cell_format
        ) if isinstance(cell_format, str) else cell_format
        write_func = getattr(self.wks, write_func) if isinstance(write_func, str) else write_func
        write_func(cell, cell_text, cell_format)

    def add_formula_column(
            self,
            df: pd.DataFrame,
            new_column_name: str,
            row_test: Callable[[DataFrameRow], bool],
            formula_func: Callable[[DataFrameRow, DataFrameIndex], str]
    ) -> pd.DataFrame:
        """
        Adds formula column at right side of given DataFrame according to row and index values

        Parameters
        ----------
        df: pd.DataFrame
        new_column_name: str
        row_test: Callable[[DataFrameRow], bool]
        formula_func: Callable[[DataFrameRow, DataFrameIndex], str]

        Returns
        -------
        df: pd.DataFrame
            Updated DataFrame object with new formula row added at right
        """

        def get_formula(row):
            """
            Passes row to ``row_test`` returning formatted Excel formula if True
            """
            if not row_test(row):
                return ""
            return formula_func(row, row.Index)

        df[new_column_name] = df.apply(
            lambda row: get_formula(row),
            axis=1
        )

        return df

    def add_list_data_validation_column(
            self,
            df: pd.DataFrame,
            new_column_name: str,
            row_test: Callable[[DataFrameRow], bool],
            *list_values: str
    ) -> pd.DataFrame:
        """
        Updates ``df`` with comma-separated list values string where ``row_test`` is True.
        Adds column header to ``data_validation_columns`` attribute for later use.

        Parameters
        ----------
        df: pd.DataFrame
        new_column_name: str
        row_test: Callable[[DataFrameRow], bool]
        *list_values: str

        Returns
        -------
        df: pd.DataFrame
        """

        # Adds data validation column to ``format_only_columns`` list to avoid writing it later
        # TODO: Figure out how to write default values to data validation list cells
        self.data_validation_columns[new_column_name] = row_test

        # Use ``row_test`` function to create mask for vectorized mapping of data validation list values
        mask = df.apply(
            lambda row: row_test(row),
            axis=1
        )

        # Map data validation values to cells where row test is True
        df[new_column_name] = np.where(
            mask,
            ",".join(list_values),
            ""
        )

        return df

    def add_right_df(
            self,
            df: pd.DataFrame,
            right_df: pd.DataFrame,
            use_index: bool = True
    ) -> pd.DataFrame:
        """
        Adds full ``right_df`` DataFrame to original ``df``

        Parameters
        ----------
        df: pd.DataFrame
        right_df: pd.DataFrame
        use_index: bool

        Returns
        -------
        df: pd.DataFrame
            Updated DataFrame object with columns from ``right_df`` added at right
        """

        if use_index:
            return df.join(right_df)

        for column_name in right_df.columns:
            # Not using index directly so first must check that columns have same length
            if right_df[column_name].shape[0] == df.shape[0]:
                df[column_name] = right_df[column_name]

        return df

    def _write_column_headers(
            self,
            df: pd.DataFrame,
            header_format: str = "bold",
            ignore_strings: List[str] = None
    ) -> None:
        """

        Parameters
        ----------
        df: pd.DataFrame
        row_number: int
        header_format: str
        ignore_strings: List[str]
            List of case-insensitive strings to skip writing as column headers (default is "index" and "unnamed")
        """
        column_headers = df.columns.tolist()
        ignore_strings = ignore_strings or ("index", "unnamed")
        columns = self.get_column_generator(
            limit=len(column_headers),
            row_number=self.current_row
        )
        column_format = self.FORMATS.get(header_format)
        for (cell, _), column_header in zip(columns, column_headers):
            if any(
                    ignore_string in column_header.lower()
                    for ignore_string in ignore_strings
            ):
                self.write_cell(
                    cell,
                    'write_string',
                    column_header,
                    column_format
                )
        self.current_row += 1

    def _set_skip_rows(
            self,
            df: pd.DataFrame,
            skip_rows: int,
            include_index: bool = False
    ) -> pd.DataFrame:
        """
        Syncs df and skip_rows so that index matches for easier manipulation

        Parameters
        ----------
        df: pd.DataFrame
        skip_rows: int

        Returns
        -------
        df: pd.DataFrame
        """
        self.current_row += skip_rows
        df = df.reset_index(drop=True)
        df.index += self.current_row
        if include_index:
            df = df.reset_index().rename(columns={"index": "_index"})
            df.index += self.current_row
        return df

    def _get_format(
            self,
            formats: List[RexcelFormat],
            row: np.ndarray,
            column_header: str,
            cell_value: Any
    ) -> Union[Any, None]:
        # First check for cell-level formats
        for cell_formatter in filter(
                lambda format: format.check_type == "cell",
                formats
        ):
            if cell_format := cell_formatter.get_format(
                    (column_header, cell_value)
            ):
                return cell_format
        # If no cell-level formats found, try column-level
        else:
            for column_formatter in filter(
                    lambda format: format.check_type == "column",
                    formats
            ):
                if column_format := column_formatter.get_format(
                        column_header
                ):
                    return column_format
            # Finally, look for row-level formats
            else:
                for row_formatter in filter(
                        lambda format: format.check_type == "row",
                        formats
                ):
                    if row_format := row_formatter.get_format(
                            row
                    ):
                        return row_format

    def write_data_validation_list(
            self,
            row: DataFrameRow,
            column_header: str,
            cell_address: str,
            cell_value: Any
    ):
        if (
                (row_test := self.data_validation_columns.get(column_header))
                and (row_test(row))
        ):
            self.wks.data_validation(
                cell_address,
                {
                    "validate": "list",
                    "source": cell_value.split(",")
                }
            )

    @staticmethod
    def _get_row_edges(df: pd.DataFrame):
        return df.index[0], df.index[-1]

    def _apply_conditional_format(
            self,
            format_range: str,
            config: dict,
            start: int,
            end: int
    ):
        format_range = format_range.format_map(
            {
                "start": start,
                "end": end
            }
        )
        self.wks.conditional_format(
            format_range,
            config
        )

    def insert_image(
            self,
            target_cell: str,
            image_file: Path,
            offset: Tuple[int, int] = None,
            scale: Tuple[int, int] = None,
            hyperlink: str = None,
            hyperlink_tip: str = None
    ):
        config = {}

        if offset is not None:
            x_offset, y_offset = offset
            config["x_offset"] = x_offset
            config["y_offset"] = y_offset

        if scale is not None:
            x_scale, y_scale = scale
            config["x_scale"] = x_scale
            config["y_scale"] = y_scale

        if hyperlink is not None:
            config["url"] = hyperlink
            if hyperlink_tip is not None:
                config["tip"] = hyperlink_tip

        self.wks.insert_image(
            target_cell,
            f"{image_file.absolute()}",
            config
        )


    def write(
            self,
            df: pd.DataFrame,
            skip_rows: int = 0,
            include_index: bool = False,
            header_format: str = "bold",
            ignore_strings: List[str] = None,
            formats: List[RexcelFormat] = None,
            conditional_formats: Dict[str, Any] = None,
            hidden_rows: List[int] = None,
            hidden_columns: List[int] = None,
            freeze_panes: Tuple[Any, Any] = None,
            hide_right_columns: str = None
    ):

        # Sync skip rows if needed
        if skip_rows:
            df = self._set_skip_rows(
                df,
                skip_rows,
                include_index=include_index
            )

        # Write column headers from ``df``
        self._write_column_headers(
            df,
            header_format=header_format,
            ignore_strings=ignore_strings
        )

        # TODO: Create better mapping to data types for rydb module
        write_funcs = {
            str: self.wks.write_string,
            bool: self.wks.write_boolean,
            int: self.wks.write_number
        }

        columns = df.columns.tolist()
        start_row, end_row = self._get_row_edges(df)

        # Iterate over rows as generator without initializing values list object
        for row in df.fillna("").values:
            row_number = self.current_row
            column_gen = self._get_column_generator(df.shape[1], row_number)
            for (column_letter, column_offset, cell_address), cell_value in zip(column_gen, row):
                cell_value = str(cell_value) if type(cell_value) not in write_funcs else cell_value

                column_header = columns[column_offset]

                # Skips columns flagged earlier as data validation lists
                if column_header in self.data_validation_columns:
                    self.write_data_validation_list(
                        row,
                        column_header,
                        cell_address,
                        cell_value
                    )
                    continue

                cell_format = self._get_format(
                    formats,
                    row,
                    column_header,
                    cell_value
                )

                cell_info = [
                    row_number,
                    column_offset,
                    cell_value,
                    cell_format
                ]

                write_func = write_funcs.get(type(cell_value))
                if cell_value and isinstance(cell_value, str) and cell_value[0] == '=':
                    write_func = self.wks.write_formula
                if cell_value == "":
                    write_func = self.wks.write_blank
                write_func(
                    *cell_info
                )

            self.current_row += 1

        # Apply conditional formats
        conditional_formats = conditional_formats or {}
        for format_range, config in conditional_formats.items():
            self._apply_conditional_format(
                format_range,
                config,
                start_row,
                end_row
            )

        for hidden_row in hidden_rows:
            self.wks.set_row(
                hidden_row,
                None,
                None,
                {'hidden': True}
            )
        for hidden_column in hidden_columns:
            self.wks.set_column(
                f"{hidden_column}:{hidden_column}",
                None,
                None,
                {'hidden': True}
            )
        if freeze_panes:
            freeze_row, freeze_column = freeze_panes
            self.wks.freeze_panes(freeze_row, freeze_column)
        if hide_right_columns is not None:
            self.wks.set_column(
                hide_right_columns,
                None,
                None,
                {
                    'hidden': True
                }
            )


class RexcelWorkbook:
    """

    """

    COLS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

    def __init__(
            self,
            output_file: Path
    ) -> None:
        self.worksheets = []
        self.new_worksheets = {}
        self.output_file = output_file

    def __enter__(self):
        self.wb = Workbook(self.output_file)
        return self

    def __exit__(self, type, value, traceback):
        self.wb.close()

    def add_format(self, config: dict):
        return self.wb.add_format(config)

    @staticmethod
    def get_column_letter(column_index: int):
        COLS = RexcelWorkbook.COLS
        if column_index < 26:
            return COLS[column_index]
        first_letter = COLS[(column_index // 26) - 1]
        second_letter = COLS[column_index % 26]
        return f"{first_letter}{second_letter}"

    def add_worksheet_by_dataframe(
            self,
            df: pd.DataFrame,
            worksheet_name: str = 'Master',
            column_widths: list = None,
            include_index: bool = False,
            format_rows: Tuple[Callable, Format] = None,
            format_columns: dict = None,
            formula_columns: dict = None,
            conditional_formatting: dict = None,
            hidden_rows: list = None,
            hidden_columns: list = None,
            header_calculations: list = None,
            skip_rows: int = 0,
            freeze_panes: Tuple[int, int] = None,
            data_validation_columns: dict = None,
            right_df: pd.DataFrame = None,
            comment_column: str = None,
            header_format: Format = None,
            hide_right_columns: str = None,

    ):
        hidden_rows = hidden_rows or []
        hidden_columns = hidden_columns or []
        header_calculations = header_calculations or []
        df = df.where(
            pd.notnull(df), ''
        )
        row_number = skip_rows
        format_test, row_format = format_rows if format_rows else (None, None)

        wks = self.wb.add_worksheet(name=worksheet_name)

        FORMATS = {
            'bold': self.wb.add_format({'bold': 1}),
            'text': self.wb.add_format({'num_format': '@'}),
            'percent': self.wb.add_format({'num_format': 9}),
            'integer': self.wb.add_format({'num_format': 1}),
            'decimal': self.wb.add_format({'num_format': 2}),
            'locked': self.wb.add_format({'locked': True})
        }

        if column_widths:
            for first, last, width in column_widths:
                wks.set_column(first, last, width)

        if header_calculations:
            for cell, write_func, cell_text, cell_format in header_calculations:
                cell_format = FORMATS.get(cell_format)
                write_func = getattr(wks, write_func)
                write_func(cell, cell_text, cell_format)

        # Create list of column headers
        columns = df.columns.tolist()

        # Add formula column headers, if present
        if formula_columns is not None:
            for formula_column in formula_columns:
                columns.append(formula_column)

        # Add data validation column headers, if present
        if data_validation_columns is not None:
            for data_validation_column in data_validation_columns:
                columns.append(data_validation_column)

        # Add right DataFrame column headers, if present
        if right_df is not None:
            columns.extend(right_df.columns.tolist())

        # Add comment column header, if present
        if comment_column is not None:
            columns.append(comment_column)

        # Write column headers
        for j, header in enumerate(columns):
            if not header or header.lower() in ('index',) or 'unnamed:' in header.lower():
                continue
            wks.write(
                f"{self.get_column_letter(j)}{row_number + 1}",
                header,
                header_format or FORMATS.get('bold')
            )

        # Create list of DataFrame values
        # Create index column, as needed
        if include_index:
            df = df.reset_index()
        rows = df.values.tolist()

        # Create list of right DataFrame values, if present
        right_rows = right_df.values.tolist() if right_df is not None else right_df

        # Increase starting row number to account for header row
        row_number += 1

        write_funcs = {
            str: wks.write_string,
            bool: wks.write_boolean,
            int: wks.write_number
        }
        for i, row in enumerate(rows):
            col_number = 0
            for j, cell in enumerate(row):
                if type(cell) not in write_funcs:
                    cell = str(cell)
                cell_info = [
                    row_number,
                    col_number,
                    cell
                ]
                if row_format is not None and format_test(row_number):
                    cell_info.append(row_format)
                if format_columns and (column_format := format_columns.get(columns[col_number])):
                    column_format = FORMATS.get(column_format)
                    cell_info.append(column_format)
                write_func = write_funcs.get(type(cell))
                if cell and isinstance(cell, str) and cell[0] == '=':
                    write_func = wks.write_formula
                if cell == "":
                    write_func = wks.write_blank
                write_func(
                    *cell_info
                )
                col_number += 1

            # Adds formula column info rendered redundant above
            if formula_columns is not None:
                for row_test, formula_format in formula_columns.values():
                    if row_test(row):
                        cell_info = [
                            row_number,
                            col_number,
                            formula_format(row, row_number)
                        ]
                        wks.write_formula(*cell_info)
                    col_number += 1
            if data_validation_columns is not None:
                for row_test, data_validation in data_validation_columns.values():
                    if row_test(row):
                        cell = f"{self.get_column_letter(col_number)}{row_number}"
                        wks.data_validation(
                            cell,
                            data_validation
                        )
                    col_number += 1
            if right_rows is not None:
                for k, cell in enumerate(right_rows[i]):
                    if type(cell) not in write_funcs:
                        cell = str(cell)
                    cell_info = [
                        row_number,
                        col_number,
                        cell
                    ]
                    write_func = write_funcs.get(type(cell))
                    if cell and isinstance(cell, str) and cell[0] == '=':
                        write_func = wks.write_formula
                    if cell == "":
                        write_func = wks.write_blank
                    write_func(
                        *cell_info
                    )
                    col_number += 1
            if comment_column is not None:
                wks.write_blank(
                    row_number,
                    col_number,
                    ""
                )
                col_number += 1
            row_number += 1
        if conditional_formatting:
            for format_range, config in conditional_formatting.items():
                if '{end}' in format_range:
                    format_range = format_range.format(end=len(rows) + 1)
                wks.conditional_format(
                    format_range,
                    config
                )
        for hidden_row in hidden_rows:
            wks.set_row(hidden_row, None, None, {'hidden': True})
        for hidden_column in hidden_columns:
            wks.set_column(f"{hidden_column}:{hidden_column}", None, None, {'hidden': True})
        if freeze_panes:
            freeze_row, freeze_column = freeze_panes
            wks.freeze_panes(freeze_row, freeze_column)
        if hide_right_columns is not None:
            wks.set_column(
                hide_right_columns,
                None,
                None,
                {
                    'hidden': True
                }
            )
        self.worksheets.append(wks)

    def add_worksheet(
            self,
            worksheet_name: str,
            column_widths: List[Tuple[int, int, int]] = None,
            image_config: dict = None
    ) -> RexcelWorksheet:
        wks = RexcelWorksheet(
            self.wb,
            worksheet_name,
            column_widths=column_widths,
            image_config=image_config
        )
        self.new_worksheets[worksheet_name] = wks
        return wks

    def new_add_worksheet_by_dataframe(
            self,
            df: pd.DataFrame,
            worksheet_name: str = 'Master',
            column_widths: list = None,
            include_index: bool = False,
            formula_columns: dict = None,
            conditional_formatting: dict = None,
            hidden_rows: list = None,
            hidden_columns: list = None,
            ad_hoc_cells: list = None,
            skip_rows: int = 0,
            freeze_panes: Tuple[int, int] = None,
            data_validation_columns: dict = None,
            right_df: pd.DataFrame = None,
            comment_column: str = None,
            header_format: Format = None,
            hide_right_columns: str = None,
            ignore_strings: List[str] = None,
            formats: list = None

    ):
        hidden_rows = hidden_rows or []
        hidden_columns = hidden_columns or []
        ad_hoc_cells = ad_hoc_cells or []

        # Register new worksheet
        wks = self.add_worksheet(
            worksheet_name=worksheet_name,
            column_widths=column_widths
        )

        # Write ad hoc cells
        for cell_config in ad_hoc_cells:
            wks.write_cell(**cell_config)

        # Add formula columns to ``df`` one by one
        # TODO: Convert these to config object
        for new_column_name, (row_test, formula_func) in formula_columns.items():
            df = wks.add_formula_column(
                df,
                new_column_name,
                row_test,
                formula_func
            )

        # Add data validation columns to ``df`` one by one
        # TODO: Currently this processes these in the same hard-coded built-ins order. Add ability to reorder columns
        data_validation_columns = data_validation_columns or {}
        for new_column_name, (row_test, config) in data_validation_columns.items():
            if list_values := config.get("source"):
                df = wks.add_list_data_validation_column(
                    df,
                    new_column_name,
                    row_test,
                    *list_values
                )

        # Add ``right_df`` data
        if right_df is not None:
            df = wks.add_right_df(df, right_df)

        # Add comment column
        if comment_column is not None:
            df[comment_column] = ""

        # Write compiled ``df`` to worksheet
        wks.write(
            df,
            skip_rows=skip_rows,
            include_index=include_index,
            header_format=header_format,
            ignore_strings=ignore_strings,
            formats=formats,
            conditional_formats=conditional_formatting,
            hidden_rows=hidden_rows,
            hidden_columns=hidden_columns,
            freeze_pandes=freeze_panes,
            hide_right_columns=hide_right_columns
        )

    def group_dfs_to_sheet(
            self,
            df: pd.DataFrame,
            columns: List[str],
            worksheet_name,
            starting_cell,
            header_format,
            skip_between: int = 1,
            column_widths=None,
            subtotal: str = None,
            image_config: dict = None,
            rate_df: pd.DataFrame = None
    ):
        df = df[columns]
        summary_groups = df.groupby(by=columns[:-1]).sum().reset_index()
        group_keys = df[columns[0]].unique().tolist()
        dfs = {
            group_key: summary_groups[
                summary_groups[columns[0]] == group_key
            ].copy()
            for group_key in group_keys
        }

        # Register new worksheet
        wks = self.add_worksheet(
            worksheet_name=worksheet_name,
            column_widths=column_widths,
            image_config=image_config
        )

        # List to collect cell addresses for subtotal cells
        subtotal_cells = []
        apply_columns = True

        # Loop over df list applying them to the worksheet
        for subtitle, df in dfs.items():
            if rate_df is not None:
                df["Rate"] = df.apply(
                    lambda x: rate_df.loc[(x["Task Type"], x["Locale"]), "Rate"],
                    axis=1
                ).fillna(0)
                df["Subtotal"] = df["Quantity"] * df["Rate"]
            current_cell = wks.simple_write_df(
                df,
                starting_cell,
                subtitle=subtitle,
                subtotal=subtotal,
                apply_columns=apply_columns
            )
            apply_columns = False
            if subtotal is not None:
                subtotal_cells.append(current_cell.pos)
            for _ in range(skip_between):
                current_cell.down()
            starting_cell = current_cell

        # Add total cell if subtotal cells have been collected
        if subtotal_cells:
            starting_cell.return_right()
            wks.write_cell(
                starting_cell.pos,
                "write_string",
                "total",
                header_format
            )
            starting_cell.right()
            total_range = ",".join(subtotal_cells)
            wks.write_cell(
                starting_cell.pos,
                "write_formula",
                f"=SUM({total_range})"
            )





RED = {
    'bg_color': '#FFC7CE',
    'font_color': '#9C0006'
}

YELLOW = {
    'bg_color': '#FFEB9C',
    'font_color': '#9C6500'
}

GREEN = {
    'bg_color': '#C6EFCE',
    'font_color': '#006100'
}

EMPTY_CELL = ""
